# Copyright 2013-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Am implementation of Reahl using SqlAlchemy Declarative.

Run 'reahl componentinfo reahl-web-declarative' for configuration information.

"""

import random
import hmac
import hashlib
import urllib.parse
from datetime import datetime, timedelta



from sqlalchemy import Column, Integer, BigInteger, LargeBinary, PickleType, String, UnicodeText, ForeignKey, DateTime
from sqlalchemy.orm import relationship, deferred

from reahl.sqlalchemysupport import Session, Base
from reahl.component.eggs import ReahlEgg
from reahl.component.config import Configuration
from reahl.component.context import ExecutionContext
from reahl.web.interfaces import UserSessionProtocol, UserInputProtocol, PersistedExceptionProtocol, PersistedFileProtocol
from reahl.web.csrf import CSRFToken
from reahl.web.fw import Url

class InvalidKeyException(Exception):
    pass

class UserSession(Base, UserSessionProtocol):
    """An implementation of :class:`reahl.web.interfaces.UserSessionProtocol` of the Reahl framework."""

    __tablename__ = 'usersession'

    id = Column(Integer, primary_key=True)
    discriminator = Column('row_type', String(40))
    __mapper_args__ = {
        'polymorphic_identity':'usersession',
        'polymorphic_on': discriminator}

    idle_lifetime = Column(Integer(), nullable=False, default=0)
    last_activity = Column(DateTime(), nullable=False, default=datetime.now)

    salt = Column(String(40), nullable=False)
    secure_salt = Column(String(40), nullable=False)

    @classmethod
    def remove_dead_sessions(cls, now=None):
        now = now or datetime.now()
        if now.minute > 0 and now.minute < 5:
            config = ExecutionContext.get_context().config
            cutoff = now - timedelta(seconds=config.web.session_lifetime)
            Session.query(cls).filter(cls.last_activity <= cutoff).delete()

    @classmethod
    def for_current_session(cls):
        return ExecutionContext.get_context().session

    @classmethod
    def initialise_web_session_on(cls, context):
        context.session = cls.get_or_create_session()

    @classmethod
    def preserve_session(cls, session):
        Session.expunge(session)

    @classmethod
    def restore_session(cls, session):
        Session.add(session)

    def __init__(self, **kwargs):
        self.generate_salt()
        self.last_activity = datetime.fromordinal(1)
        self.set_idle_lifetime(False)
        super().__init__(**kwargs)

    def get_csrf_token(self):
        key = ExecutionContext.get_context().config.web.csrf_key
        message = self.as_key()
        return CSRFToken(value=hmac.new(key.encode('utf-8'), msg=message.encode('utf-8'), digestmod=hashlib.sha1).hexdigest())

    def is_secured(self):
        context = ExecutionContext.get_context()
        return self.is_within_timeout(context.config.web.idle_secure_lifetime) \
               and context.request.scheme == 'https' \
               and self.secure_cookie_is_valid()

    def is_active(self):
        return self.is_within_timeout(self.idle_lifetime)

    def is_within_timeout(self, timeout):
        return self.last_activity + timedelta(seconds=timeout) > datetime.now()

    def set_last_activity_time(self):
        self.last_activity = datetime.now()

    def set_idle_lifetime(self, use_max):
        config = ExecutionContext.get_context().config
        self.idle_lifetime = config.web.idle_lifetime_max if use_max else config.web.idle_lifetime

    @classmethod
    def from_key(cls, key):
        try:
            web_session_id, salt = key.split(':')
        except ValueError:
            raise InvalidKeyException()
        sessions = Session.query(cls).filter_by(id=web_session_id)
        if sessions.count() == 1 and sessions.one().salt == salt:
            return sessions.one()
        raise InvalidKeyException()

    def as_key(self):
        Session.flush() # To make sure .id is populated
        return '%s:%s' % (str(self.id), self.salt)

    def secure_cookie_is_valid(self):
        context = ExecutionContext.get_context()
        try:
            salt = context.request.cookies[context.config.web.secure_key_name]
            return self.secure_salt == salt
        except KeyError:
            return False

    @classmethod
    def get_or_create_session(cls):
        web_session = None
        session_key = cls.get_session_key()
        if session_key:
            try:
                web_session = cls.from_key(session_key)
            except InvalidKeyException:
                pass

        if not web_session:
            web_session = cls()
            Session.add(web_session)

        return web_session

    @classmethod
    def get_session_key(cls):
        context = ExecutionContext.get_context()
        try:
            raw_cookie = context.request.cookies[context.config.web.session_key_name]
            return urllib.parse.unquote(raw_cookie)
        except KeyError:
            return None

    def set_session_key(self, response):
        context = ExecutionContext.get_context()
        session_cookie = self.as_key()
        response.set_cookie(context.config.web.session_key_name, urllib.parse.quote(session_cookie), path='/', samesite='Strict')
        if self.is_secured():
            response.set_cookie(context.config.web.secure_key_name, urllib.parse.quote(self.secure_salt), secure=True, path='/',
                                max_age=context.config.web.idle_secure_lifetime, samesite='Strict')

    def generate_salt(self):
        alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZqwertyuiopasdfghjklzxcvbnm0123456789'
        self.salt = ''.join([random.choice(alphabet) for x in list(range(40))])
        self.secure_salt = ''.join([random.choice(alphabet) for x in list(range(40))])        

    def get_interface_locale(self):
        context = ExecutionContext.get_context()
        if not hasattr(context, 'request'):
            return 'en_gb'

        url = Url.get_current_url()
        possible_locale, path = url.get_locale_split_path()
        supported_locales = ReahlEgg.get_languages_supported_by_all(context.config.reahlsystem.root_egg)
        if possible_locale:
            if possible_locale in supported_locales:
                return possible_locale
        return context.config.web.default_url_locale


class SessionData(Base):
    __tablename__ = 'sessiondata'

    id = Column(Integer, primary_key=True)
    discriminator = Column('row_type', String(40))
    __mapper_args__ = {
        'polymorphic_identity': 'sessiondata',
        'polymorphic_on': discriminator}
    
    web_session_id = Column(Integer, ForeignKey('usersession.id', ondelete='CASCADE'), index=True)
    web_session = relationship(UserSession)

    view_path = Column(UnicodeText, nullable=False)
    ui_name = Column(UnicodeText, nullable=False)
    channel_name = Column(UnicodeText, nullable=True)

    @classmethod
    def clear_for_form(cls, form):
        for stale in cls.find_for(form.view, form=form).all():
            Session.delete(stale)

    @classmethod
    def clear_for_view(cls, view):
        for stale in cls.find_for(view, form=None).all():
            Session.delete(stale)

    @classmethod
    def clear_all_view_data(cls, view):
        web_session = ExecutionContext.get_context().session
        items = Session.query(cls).filter_by(web_session=web_session, view_path=view.full_path, ui_name=view.user_interface.name)
        for stale in items:
            Session.delete(stale)
        
    @classmethod
    def find_for(cls, view, form=None):
        assert (not form) or (form.view is view)
        web_session = ExecutionContext.get_context().session
        channel_name = form.channel_name if form else None
        return Session.query(cls).filter_by(web_session=web_session, view_path=view.full_path, ui_name=view.user_interface.name, channel_name=channel_name)

    @classmethod
    def save_for(cls, view, form=None, **kwargs):
        assert (not form) or (form.view is view)
        channel_name = form.channel_name if form else None
        web_session = ExecutionContext.get_context().session
        instance = cls(web_session=web_session, view_path=view.full_path, ui_name=view.user_interface.name, channel_name=channel_name, **kwargs)
        Session.add(instance)
        return instance

    __hash__ = None
    def __eq__(self, other):
        return self.web_session == other.web_session and \
               self.ui_name == other.ui_name and \
               self.channel_name == other.channel_name

    def __neq__(self, other):
        return not self.__eq__(other)


class UserInput(SessionData, UserInputProtocol):
    """An implementation of :class:`reahl.web.interfaces.UserInputProtocol`. It represents
       a value that was input by a user."""
    __tablename__ = 'userinput'
    __mapper_args__ = {'polymorphic_identity': 'userinput'}
    id = Column(Integer, ForeignKey('sessiondata.id', ondelete='CASCADE'), primary_key=True)

    key = Column(UnicodeText, nullable=False)
    value = Column(UnicodeText, nullable=True)

    __hash__ = None
    def __eq__(self, other):
        return super().__eq__(other) and \
               self.key == other.key and \
               self.value == other.value

    @classmethod
    def get_previously_entered_for_form(cls, form, input_name, entered_input_type):
        return cls.get_previously_saved_for(form.view, form, input_name, entered_input_type)

    @classmethod
    def get_previously_saved_for(cls, view, form, key, value_type):
        query = cls.find_for(view, form=form).filter_by(key=key)
        if query.count() == 0:
            return None

        values = [i.value for i in query.all()]

        if value_type is str:
            assert len(values) == 1, 'There are %s saved values for "%s", but there should only be one' % (len(values), key)
            return values[0]
        elif value_type is list:
            if len(values) == 1 and values[0] is None:
                return []
            return values
        else:
            assert None, 'Cannot persist values of type: %s' % value_type

    @classmethod
    def save_input_value_for_form(cls, form, input_name, value, entered_input_type):
        cls.save_value_for(form.view, form, input_name, value, entered_input_type)

    @classmethod
    def save_value_for(cls, view, form, key, value, value_type):
        for i in cls.find_for(view, form=form).filter_by(key=key):
            Session.delete(i)

        if value_type is str:
            assert isinstance(value, str), 'Cannot handle the value: ' + str(value)
            cls.save_for(view, form=form, key=key, value=value)
        elif value_type is list:
            assert all([isinstance(i, str) for i in value]), 'Cannot handle the value: ' + str(value)
            if value:
                for i in value:
                    cls.save_for(view, form=form, key=key, value=i)
            else:
                cls.save_for(view, form=form, key=key, value=None)
        else:
            assert None, 'Cannot persist values of type: %s' % value_type

    @classmethod
    def get_persisted_for_view(cls, view, key, value_type):
        return cls.get_previously_saved_for(view, None, key, value_type)

    @classmethod
    def add_persisted_for_view(cls, view, key, value, value_type):
        cls.save_value_for(view, None, key, value, value_type)

    @classmethod
    def remove_persisted_for_view(cls, view, key):
        for previous in cls.find_for(view, form=None).filter_by(key=key):
            Session.delete(previous)


class PersistedException(SessionData, PersistedExceptionProtocol):
    """An implementation of :class:`reahl.web.interfaces.PersistedExceptionProtocol`. It represents
       an Exception that was raised upon a user interaction."""
    __tablename__ = 'persistedexception'
    __mapper_args__ = {'polymorphic_identity': 'persistedexception'}
    id = Column(Integer, ForeignKey('sessiondata.id', ondelete='CASCADE'), primary_key=True)

    exception = Column(PickleType, nullable=False)
    input_name = Column(UnicodeText)

    def __eq__(self, other):
        return super().__eq__(other) and \
               self.exception == other.exception

    @classmethod
    def save_exception_for_form(cls, form, **kwargs):
        return cls.save_for(form.view, form=form, **kwargs)

    @classmethod
    def for_input(cls, form, input_name):
        return super().find_for(form.view, form=form).filter_by(input_name=input_name)

    @classmethod
    def clear_for_form_except_inputs(cls, form):
        for stale in cls.for_input(form, None).all():
            Session.delete(stale)

    @classmethod
    def clear_for_all_inputs(cls, form):
        for e in super().find_for(form.view, form=form).filter(cls.input_name is not None):
            Session.delete(e)

    @classmethod
    def get_exception_for_form(cls, form):
        persisted = cls.for_input(form, None)
        if persisted.count() == 1:
            return persisted.one().exception
        return None

    @classmethod
    def get_exception_for_input(cls, form, input_name):
        persisted = cls.for_input(form, input_name)
        if persisted.count() == 1:
            return persisted.one().exception
        return None


class PersistedFile(SessionData, PersistedFileProtocol):
    """An implementation of :class:`reahl.web.interfaces.PersistedFileProtocol`. It represents
       a file that was input by a user."""
    __tablename__ = 'persistedfile'
    __mapper_args__ = {'polymorphic_identity': 'persistedfile'}
    id = Column(Integer, ForeignKey('sessiondata.id', ondelete='CASCADE'), primary_key=True)

    input_name = Column(UnicodeText, nullable=False)
    filename = Column(UnicodeText, nullable=False)
    file_data = deferred(Column(LargeBinary, nullable=False)) 
    mime_type = Column(UnicodeText, nullable=False)
    size = Column(BigInteger, nullable=False)

    def __eq__(self, other):
        return super().__eq__(other) and \
               self.filename == other.filename and \
               self.input_name == other.input_name

    @property
    def file_obj(self):
        class LazyFileObj:
            def __init__(self, persisted_file):
                self.persisted_file = persisted_file
                self.position = 0
            def read(self):
                try:
                    return self.persisted_file.file_data[self.position:]
                finally:
                    self.position = len(self.persisted_file.file_data)-1
            def seek(self, position):
                self.position = position
        return LazyFileObj(self)

    @classmethod
    def get_persisted_for_form(cls, form, input_name):
        query = cls.find_for(form.view, form=form).filter_by(input_name=input_name)
        return query.all()

    @classmethod
    def add_persisted_for_form(cls, form, input_name, uploaded_file):
        filename = uploaded_file.filename
        file_data = uploaded_file.contents
        mime_type = uploaded_file.mime_type
        size = uploaded_file.size
        return cls.save_for(form.view, form=form, input_name=input_name, filename=filename, file_data=file_data, mime_type=mime_type, size=size)

    @classmethod
    def remove_persisted_for_form(cls, form, input_name, filename):
        persisted = cls.find_for(form.view, form=form).filter_by(input_name=input_name).filter_by(filename=filename)
        for persisted_file in persisted.all():
            Session.delete(persisted_file)

    @classmethod
    def is_uploaded_for_form(cls, form, input_name, filename):
        return PersistedFile.find_for(form.view, form=form).filter_by(input_name=input_name).filter_by(filename=filename).count() == 1
    

class WebDeclarativeConfig(Configuration):
    filename = 'web.webdeclarative.config.py'
    config_key = 'web.webdeclarative'

    def do_injections(self, config):
        config.web.session_class = UserSession
        config.web.persisted_exception_class = PersistedException
        config.web.persisted_userinput_class = UserInput
        config.web.persisted_file_class = PersistedFile





# Copyright 2013-2020 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from datetime import datetime, timedelta

import http.cookies
import urllib.parse

from sqlalchemy import Column, ForeignKey, Integer
from webob import Response

from reahl.tofu import scenario, Fixture, uses
from reahl.stubble import stubclass, EmptyStub
from reahl.tofu.pytestsupport import with_fixtures

from reahl.component.context import ExecutionContext
from reahl.sqlalchemysupport import Session, Base, session_scoped
from reahl.webdeclarative.webdeclarative import UserSession, SessionData, UserInput

from reahl.web.ui import Form
from reahl.web_dev.fixtures import WebFixture
from reahl.sqlalchemysupport_dev.fixtures import SqlAlchemyFixture
from reahl.domain_dev.fixtures import PartyAccountFixture


@with_fixtures(WebFixture)
def test_session_active_state(web_fixture):
    """The session is active if the last user interaction was in the last idle_lifetime """
    fixture = web_fixture
    user_session = fixture.context.session
    config = fixture.config

    # Case: recent interaction
    user_session.set_last_activity_time()
    assert user_session.is_active()

    # Case: last interaction not recent
    user_session.last_activity = datetime.now() - timedelta(seconds=user_session.idle_lifetime+10)
    assert not user_session.is_active()


@uses(web_fixture=WebFixture)
class SecureScenarios(Fixture):

    @scenario
    def secure(self):
        self.scheme = 'https'
        self.last_activity = datetime.now()
        self.secure_cookie = self.web_fixture.context.session.secure_salt
        self.expect_secure = True

    @scenario
    def insecure_scheme(self):
        self.secure()
        self.scheme = 'http'
        self.expect_secure = False

    @scenario
    def old_interaction(self):
        self.secure()
        self.last_activity = datetime.now() - timedelta(seconds=self.web_fixture.config.web.idle_secure_lifetime+10)
        self.expect_secure = False

    @scenario
    def bad_cookie(self):
        self.secure()
        self.secure_cookie = 'bad cookie value'
        self.expect_secure = False


@with_fixtures(WebFixture, SecureScenarios)
def test_session_secure_state(web_fixture, secure_scenarios):
    """The session is only secured when used over https, the secure cookie is set correctly,
       and the last interaction is within idle_secure_lifetime"""
    fixture = secure_scenarios

    user_session = web_fixture.context.session
    config = web_fixture.config
    context = web_fixture.context
    assert config.web.idle_secure_lifetime < config.web.idle_lifetime
    assert config.web.idle_lifetime < config.web.idle_lifetime_max

    web_fixture.request.scheme = fixture.scheme
    user_session.last_activity = fixture.last_activity
    context.request.cookies[context.config.web.secure_key_name] = fixture.secure_cookie
    assert user_session.is_secured() is fixture.expect_secure


@with_fixtures(SqlAlchemyFixture, WebFixture)
def test_setting_cookies_on_response(sql_alchemy_fixture, web_fixture):
    """How UserSession sets session and secure cookies in the response."""
    fixture = web_fixture
    @stubclass(UserSession)
    class UserSessionStub(UserSession):
        __tablename__ = 'usersessionstub'
        __mapper_args__ = {'polymorphic_identity': 'usersessionstub'}
        id = Column(Integer, ForeignKey('usersession.id'), primary_key=True)

        secured = False
        def is_secured(self):
            return self.secured

    with sql_alchemy_fixture.persistent_test_classes(UserSessionStub):
        user_session = UserSessionStub()

        class ResponseStub(Response):
            @property
            def cookies(self):
                cookies = http.cookies.SimpleCookie()
                for header, value in self.headerlist:
                    if header == 'Set-Cookie':
                        cookies.load(value)
                return cookies

        # Case: with an unsecured session, set only the session cookie
        user_session.secured = False
        response = ResponseStub()

        user_session.set_session_key(response)

        session_cookie = response.cookies[fixture.config.web.session_key_name]
        assert session_cookie.value == urllib.parse.quote(user_session.as_key())
        assert session_cookie['path'] == '/'
        assert not session_cookie['max-age']
        #assert 'httponly' in session_cookie

        assert fixture.config.web.secure_key_name not in response.cookies


        # Case: with an secured session, set the session cookie and the secure cookie
        user_session.secured = True
        response = ResponseStub()

        user_session.set_session_key(response)

        assert fixture.config.web.session_key_name in response.cookies

        secure_cookie = response.cookies[fixture.config.web.secure_key_name]
        assert user_session.secure_salt == secure_cookie.value
        assert secure_cookie['path'] == '/'
        assert secure_cookie['max-age'] == '%s' % fixture.config.web.idle_secure_lifetime
        assert 'secure' in secure_cookie
        #assert 'httponly' in secure_cookie


@with_fixtures(WebFixture)
def test_reading_cookies_on_initialising_a_session(web_fixture):
    fixture = web_fixture

    # Case: session cookie not set in Request
    UserSession.initialise_web_session_on(fixture.context)
    assert not fixture.context.session.is_active()
    assert not fixture.context.session.is_secured()

    # Case: session cookie set in Request
    fixture.context.session = None
    user_session = UserSession()
    user_session.set_last_activity_time()
    Session.add(user_session)

    fixture.request.headers['Cookie'] = 'reahl=%s' % user_session.as_key()
    UserSession.initialise_web_session_on(fixture.context)

    assert fixture.context.session is user_session
    assert fixture.context.session.is_active()
    assert not fixture.context.session.is_secured()

    # Case: session cookie set, secure cookie also set in Request, https
    fixture.request.scheme = 'https'
    fixture.context.session = None
    user_session = UserSession()
    user_session.set_last_activity_time()
    Session.add(user_session)

    fixture.request.headers['Cookie'] = 'reahl=%s , reahl_secure=%s' % \
                                        (user_session.as_key(), user_session.secure_salt)
    UserSession.initialise_web_session_on(fixture.context)

    assert fixture.context.session is user_session
    assert fixture.context.session.is_active()
    assert fixture.context.session.is_secured()

    # Case: session cookie set, secure cookie also set in Request, http
    fixture.request.scheme = 'http'
    fixture.context.session = None
    user_session = UserSession()
    user_session.set_last_activity_time()
    Session.add(user_session)
    fixture.request.headers['Cookie'] = 'reahl=%s , reahl_secure=%s' % \
                                        (user_session.as_key(), user_session.secure_salt)

    UserSession.initialise_web_session_on(fixture.context)

    assert fixture.context.session is user_session
    assert fixture.context.session.is_active()
    assert not fixture.context.session.is_secured()


@with_fixtures(WebFixture)
def test_session_data_disappears_when_session_does(web_fixture):
    """When a UserSession is deleted, all associated SessionData disappear as well."""
    fixture = web_fixture

    UserSession.initialise_web_session_on(fixture.context)
    user_session = fixture.context.session
    ui_name = 'user_interface'
    channel_name = 'channel'

    session_data = SessionData(web_session=user_session, view_path='/', ui_name=ui_name, channel_name=channel_name)
    Session.add(session_data)
    Session.flush()

    Session.delete(user_session)

    assert Session.query(SessionData).filter_by(id=session_data.id).count() == 0
    assert Session.query(UserSession).filter_by(id=user_session.id).count() == 0


@with_fixtures(WebFixture)
def test_session_keeps_living(web_fixture):
    """When SessionData is deleted, the associated UserSession is not affected."""
    fixture = web_fixture

    UserSession.initialise_web_session_on(fixture.context)
    user_session = fixture.context.session
    ui_name = 'user_interface'
    channel_name = 'channel'

    session_data = SessionData(web_session=user_session, view_path='/', ui_name=ui_name, channel_name=channel_name)
    Session.add(session_data)
    Session.flush()

    Session.delete(session_data)

    assert Session.query(SessionData).filter_by(id=session_data.id).count() == 0
    assert Session.query(UserSession).filter_by(id=user_session.id).one() is user_session


class InputScenarios(Fixture):
    @scenario
    def text(self):
        self.entered_input_type = str
        self.entered_input = 'some value to save'
        self.empty_entered_input = ''

    @scenario
    def lists(self):
        self.entered_input_type = list
        self.entered_input = ['one', 'two']
        self.empty_entered_input = []


@with_fixtures(WebFixture, PartyAccountFixture, InputScenarios)
def test_persisting_input(web_fixture, party_account_fixture, input_scenarios):
    """UserInput can persist and find user input entered as a string or a list of strings (useful for things
       like multiple select boxes or multiple checkboxes with the same name).
    """
    @stubclass(Form)
    class FormStub:
        view = web_fixture.view
        user_interface = EmptyStub(name='myui')
        channel_name = 'myform'

    fixture = input_scenarios
    form = FormStub()

    # If never persisted, getting it returns None
    previously_entered = UserInput.get_previously_entered_for_form(form, 'aninput', fixture.entered_input_type)
    assert previously_entered is None

    # Once persisted, getting it returns what you persisted
    UserInput.save_input_value_for_form(form, 'aninput', fixture.entered_input, fixture.entered_input_type)
    previously_entered = UserInput.get_previously_entered_for_form(form, 'aninput', fixture.entered_input_type)
    assert previously_entered == fixture.entered_input

    # Persisting empty values returns them correctly, not None (None means was not entered previously)
    UserInput.clear_for_form(form)
    previously_entered = UserInput.get_previously_entered_for_form(form, 'aninput', fixture.entered_input_type)
    assert previously_entered is None

    UserInput.save_input_value_for_form(form, 'aninput', fixture.empty_entered_input, fixture.entered_input_type)
    previously_entered = UserInput.get_previously_entered_for_form(form, 'aninput', fixture.entered_input_type)
    assert previously_entered == fixture.empty_entered_input


class SessionScopedFixture(Fixture):

    def create_user_session(self):
        user_session = UserSession()
        Session.add(user_session)
        ExecutionContext.get_context().session = user_session
        return user_session

    @scenario
    def remove_session_scoped(self):
        self.expected_user_session_after_delete = 1
        self.cascade = False

    @scenario
    def remove_user_session(self):
        self.expected_user_session_after_delete = 0
        self.cascade = True


@with_fixtures(SqlAlchemyFixture, SessionScopedFixture)
def test_cascade_removal_of_user_session(sql_alchemy_fixture, session_scoped_fixture):
    """
        If a user session is deleted, all session scoped objects are deleted as well.
    """

    fixture = session_scoped_fixture

    @session_scoped
    class MySessionScoped(Base):
        __tablename__ = 'my_session_scoped'
        id = Column(Integer, primary_key=True)

    with sql_alchemy_fixture.persistent_test_classes(MySessionScoped):
        user_session = fixture.create_user_session()
        assert Session.query(MySessionScoped).count() == 0

        session_object = MySessionScoped.for_current_session()
        assert Session.query(MySessionScoped).one() is session_object
        assert session_object.user_session is user_session

        if fixture.cascade:
            Session.delete(user_session)
        else:
            Session.delete(session_object)
        Session.flush()

        assert Session.query(MySessionScoped).count() == 0
        assert Session.query(UserSession).count() == fixture.expected_user_session_after_delete



from setuptools import setup, Command
class InstallTestDependencies(Command):
    user_options = []
    def run(self):
        import sys
        import subprocess
        if self.distribution.tests_require: subprocess.check_call([sys.executable, "-m", "pip", "install", "-q"]+self.distribution.tests_require)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name='reahl-web-declarative',
    version='6.0.0',
    description='An implementation of Reahl persisted classes using SqlAlchemy.',
    long_description='Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nSome core elements of Reahl can be implemented for use with different persistence technologies. This is such an implementation based on SqlAlchemy. ',
    url='http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.webdeclarative', 'reahl.webdeclarative_dev'],
    py_modules=[],
    include_package_data=True,
    namespace_packages=['reahl'],
    install_requires=['reahl-sqlalchemysupport>=6.0,<6.1', 'reahl-web>=6.0,<6.1', 'reahl-component>=6.0,<6.1'],
    setup_requires=['pytest-runner', 'reahl-component-metadata', 'setuptools >= 51.0.0', 'setuptools-git >= 1.1', 'toml', 'wheel'],
    tests_require=['WebOb>=1.8,<1.8.999', 'pytest>=3.0', 'reahl-tofu>=6.0,<6.1', 'reahl-stubble>=6.0,<6.1', 'reahl-dev>=6.0,<6.1', 'reahl-webdev>=6.0,<6.1', 'reahl-browsertools>=6.0,<6.1', 'reahl-domain>=6.0,<6.1', 'reahl-postgresqlsupport>=6.0,<6.1'],
    extras_require={},
    cmdclass={'install_test_dependencies': InstallTestDependencies}
)



# Copyright 2014-2022 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.




from sqlalchemy import Column, String, UnicodeText, Integer, Text, PrimaryKeyConstraint, LargeBinary, BigInteger
from alembic import op

from reahl.sqlalchemysupport.elixirmigration import MigrateElixirToDeclarative
from reahl.sqlalchemysupport import fk_name, ix_name

from reahl.component.migration import Migration
from reahl.component.context import ExecutionContext


class CreateDatabase(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql')
        self.schedule('alter', op.create_table, 'sessiondata',
                      Column('id', Integer(), nullable=False),
                      Column('web_session_id', Integer(), nullable=True),
                      Column('region_name', Text(), nullable=False),
                      Column('channel_name', Text(), nullable=False),
                      Column('row_type', String(length=40), nullable=True),
                      PrimaryKeyConstraint('id', name='sessiondata_pkey')
                      )
        self.schedule('create_fk', op.create_foreign_key, 'sessiondata_web_session_id_fk', 'sessiondata',
                      'usersession', ['web_session_id'], ['id'], ondelete='CASCADE')

        # self.schedule('indexes', op.create_index, 'sessiondata_id_seq', 'sessiondata', ['id'])
        self.schedule('indexes', op.create_index, 'ix_sessiondata_web_session_id', 'sessiondata', ['web_session_id'], unique=False)

        self.schedule('alter', op.create_table, 'webusersession',
                      Column('usersession_id', Integer(), nullable=False),
                      Column('salt', String(length=40), nullable=False),
                      Column('secure_salt', String(length=40), nullable=False),
                      PrimaryKeyConstraint('usersession_id', name='webusersession_pkey')
                      )
        self.schedule('create_fk', op.create_foreign_key, 'webusersession_usersession_id_fkey', 'webusersession',
                      'usersession', ['usersession_id'], ['id'], ondelete='CASCADE')

        self.schedule('alter', op.create_table, 'persistedexception',
                      Column('sessiondata_id', Integer(), nullable=False),
                      Column('exception', LargeBinary(), nullable=False),
                      Column('input_name', Text(), nullable=True),
                      PrimaryKeyConstraint('sessiondata_id', name='persistedexception_pkey')
                      )
        self.schedule('create_fk', op.create_foreign_key, 'persistedexception_sessiondata_id_fkey', 'persistedexception',
                      'sessiondata', ['sessiondata_id'], ['id'], ondelete='CASCADE')

        self.schedule('alter', op.create_table, 'persistedfile',
                      Column('sessiondata_id', Integer(), nullable=False),
                      Column('input_name', Text(), nullable=False),
                      Column('filename', Text(), nullable=False),
                      Column('file_data', LargeBinary(), nullable=False),
                      Column('content_type', Text(), nullable=False),
                      Column('size', BigInteger(), nullable=False),
                      PrimaryKeyConstraint('sessiondata_id', name='persistedfile_pkey')
                      )
        self.schedule('create_fk', op.create_foreign_key, 'persistedfile_sessiondata_id_fkey', 'persistedfile',
                      'sessiondata', ['sessiondata_id'], ['id'], ondelete='CASCADE')

        self.schedule('alter', op.create_table, 'userinput',
                      Column('sessiondata_id', Integer(), nullable=False),
                      Column('key', Text(), nullable=False),
                      Column('value', Text(), nullable=False),
                      PrimaryKeyConstraint('sessiondata_id', name='userinput_pkey')
                      )
        self.schedule('create_fk', op.create_foreign_key, 'userinput_sessiondata_id_fkey', 'userinput',
                      'sessiondata', ['sessiondata_id'], ['id'], ondelete='CASCADE')


class RenameRegionToUi(Migration):

    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql')
        self.schedule('alter', op.alter_column, 'sessiondata', 'region_name', new_column_name='ui_name')


class ElixirToDeclarativeWebDeclarativeChanges(MigrateElixirToDeclarative):
    def schedule_upgrades(self):
        super().schedule_upgrades()
        self.replace_elixir()

    def rename_primary_key_constraints(self):
        self.rename_pk('sessiondata', ['id'])

    def rename_foreign_keys_constraints(self):
        self.recreate_foreign_key_constraint('sessiondata_web_session_id', 'sessiondata', 'web_session_id', 'webusersession', 'id', ondelete='CASCADE')

    def change_inheriting_table_ids(self):
        for table_name, old_id_column_name, inheriting_table_name in [
                              ('webusersession', 'usersession_id', 'usersession'),
                              ('persistedexception', 'sessiondata_id', 'sessiondata'),
                              ('persistedfile', 'sessiondata_id', 'sessiondata'),
                              ('userinput', 'sessiondata_id', 'sessiondata')
                            ]:
            self.change_inheriting_table(table_name, old_id_column_name, inheriting_table_name)

    def replace_elixir(self):
        # reahl-declarative is new, and replaces reahl-elixir-impl
        orm_control = ExecutionContext.get_context().system_control.orm_control
        self.schedule('cleanup', orm_control.remove_schema_version_for, egg_name='reahl-web-elixirimpl', fail_if_not_found=False)


class MergeWebUserSessionToUserSession(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql')
        self.schedule('drop_pk', op.drop_index, ix_name('usersession', 'account_id'))
        self.schedule('alter', op.drop_column, 'usersession', 'account_id')
        self.schedule('alter', op.add_column, 'usersession', Column('salt', String(40), nullable=False))
        self.schedule('alter', op.add_column, 'usersession', Column('secure_salt', String(40), nullable=False))
        self.schedule('alter', op.drop_table, 'webusersession')
        self.schedule('data', op.execute, 'delete from usersession')

        self.schedule('drop_fk', op.drop_constraint, fk_name('sessiondata', 'web_session_id', 'webusersession'), 'sessiondata')
        self.schedule('create_fk', op.create_foreign_key, fk_name('sessiondata', 'web_session_id', 'usersession'), 'sessiondata',
                      'usersession', ['web_session_id'], ['id'], ondelete='CASCADE')


class RenameContentType(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql')
        self.schedule('alter', op.add_column, 'persistedfile', Column('mime_type', UnicodeText, nullable=False))
        self.schedule('data', op.execute, 'update persistedfile set mime_type=content_type')
        self.schedule('cleanup', op.drop_column, 'persistedfile', 'content_type')


class AllowNullUserInputValue(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql')
        self.schedule('alter', op.alter_column, 'userinput', 'value', existing_nullable=False, nullable=True)


class AddViewPathToSessionData(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql', 'mysql')
        self.schedule('alter', op.alter_column, 'sessiondata', 'channel_name', existing_nullable=False, nullable=True, existing_type=UnicodeText)
        self.schedule('alter', op.add_column, 'sessiondata', Column('view_path', UnicodeText, nullable=False))


class AddPolimorphicEntityName(Migration):
    def schedule_upgrades(self):
        self.orm_control.assert_dialect(self, 'postgresql', 'mysql')
        self.schedule('data', op.execute, 'update sessiondata set row_type=\'sessiondata\' where row_type is null')
        



import os

from reahl.sqlalchemysupport import SqlAlchemyControl

reahlsystem.connection_uri = os.environ.get('REAHL_TEST_CONNECTION_URI', 'postgresql:///reahl')
reahlsystem.root_egg = 'reahl-web-declarative'
#reahlsystem.connection_uri = 'sqlite://'
reahlsystem.orm_control = SqlAlchemyControl(echo=False)
#reahlsystem.debug = True



